##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: testclient.py 90845 2008-09-04 21:06:25Z mgedmin $
"""
__docformat__ = "reStructuredText"

import sys
import optparse
import textwrap

from keas.kmi.facility import LocalKeyManagementFacility


def ping(kmf):
    client = kmf.clientClass(kmf.url)
    print client.fullStatus
    print
    print client.contents


def new_key(kmf):
    sys.stdout.write(kmf.generate())


def read_kek(kekfile):
    try:
        return file(kekfile, 'rb').read()
    except IOError, e:
        print >> sys.stderr, "Could not read key encrypting key from %s" % kekfile
        print >> sys.stderr, e
        sys.exit(1)


def read_data(filename=None):
    if not filename:
        return sys.stdin.read()
    else:
        try:
            return file(filename, 'rb').read()
        except IOError, e:
            print >> sys.stderr, "Could not read %s" % filename
            print >> sys.stderr, e
            sys.exit(1)


def get_key(kmf, kekfile):
    key_encrypting_key = read_kek(kekfile)
    key = kmf.getEncryptionKey(key_encrypting_key)
    sys.stdout.write(key)


def encrypt(kmf, kekfile, filename=None):
    key_encrypting_key = read_kek(kekfile)
    data = read_data(filename)
    encrypted = kmf.encrypt(key_encrypting_key, data)
    sys.stdout.write(encrypted)


def decrypt(kmf, kekfile, filename=None):
    key_encrypting_key = read_kek(kekfile)
    data = read_data(filename)
    decrypted = kmf.decrypt(key_encrypting_key, data)
    sys.stdout.write(decrypted)


def main():
    parser = optparse.OptionParser(textwrap.dedent("""\
                usage: %prog URL
                            see if the server is alive

                       %prog URL -n > key.txt
                            generate a new key and key encrypting key

                       %prog URL -e key.txt data.txt > encrypted.txt
                            encrypt data

                       %prog URL -d key.txt encrytped.txt > data.txt
                            decrypt data

                       %prog URL -g key.txt > secretkey.bin
                            get the secret encryption key
                """.rstrip()),
                description="Client for a Key Management Server.")
    parser.add_option('-n', '--new',
                      help='generate a new key',
                      action='store_const', dest='action',
                      const=new_key)
    parser.add_option('-g', '--get-key',
                      help='get key',
                      action='store_const', dest='action',
                      const=get_key)
    parser.add_option('-e', '--encrypt',
                      help='encrypt data',
                      action='store_const', dest='action',
                      const=encrypt)
    parser.add_option('-d', '--decrypt',
                      help='decrypt data',
                      action='store_const', dest='action',
                      const=decrypt)
    opts, args = parser.parse_args()
    if not opts.action:
        opts.action = ping
    if not args:
        parser.error('please specify the KMS server URL')

    url = args.pop(0)
    kmf = LocalKeyManagementFacility(url)

    try:
        opts.action(kmf, *args)
    except TypeError:
        parser.error('incorrect number of arguments')



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Automatic ZODB installation

$Id: db.py 90781 2008-09-04 05:59:31Z srichter $
"""
__docformat__ = "reStructuredText"

import transaction
from zope.component import adapter
from zope.app.appsetup.interfaces import IDatabaseOpenedEvent
from zope.app.appsetup.bootstrap import getInformationFromEvent
from zope.app.publication.zopepublication import ZopePublication

from keas.kmi.facility import KeyManagementFacility
from keas.kmi.interfaces import IKeyManagementFacility


@adapter(IDatabaseOpenedEvent)
def bootstrapKeyManagementFacility(event):
    """Installs KeyManagementFacility as the root object of the DB."""
    db, connection, root, root_object = getInformationFromEvent(event)
    if root_object is None:
        root[ZopePublication.root_name] = KeyManagementFacility()
        transaction.commit()
    elif not IKeyManagementFacility.providedBy(root_object):
        raise RuntimeError('Your database root object is not a key management'
                           ' facility.  Remove your Data.fs and try again.')
    connection.close()




"""
WSGI application for the Key Management Server.
"""

import os

import zope.app.appsetup
from zope.app.wsgi import getWSGIApplication


def application_factory(global_conf, conf='kms.conf', **local_conf):
    configfile = os.path.join(global_conf['here'], conf)
    application = getWSGIApplication(configfile)
    return application




"""Encrypted persistent objects

$Id: persistent.py 90781 2008-09-04 05:59:31Z srichter $
"""
from __future__ import absolute_import
import cPickle
import cStringIO
import persistent
import persistent.wref
from zope.component import getUtility
from keas.kmi.interfaces import IEncryptionService
from keas.kmi.interfaces import IKeyHolder


class EncryptedPersistent(persistent.Persistent):
    """A persistent object that is stored in encrypted form."""

    def __getstate__(self):
        state = super(EncryptedPersistent, self).__getstate__()
        return encrypt_state(state)

    def __setstate__(self, encrypted_state):
        state = decrypt_state(encrypted_state)
        super(EncryptedPersistent, self).__setstate__(state)


def encrypt_state(state):
    key = getUtility(IKeyHolder).key
    service = getUtility(IEncryptionService)
    stringified_data, persistent_refs = pickle_nonpersistent(state)
    encrypted_data = service.encrypt(key, stringified_data)
    return encrypted_data, persistent_refs


def decrypt_state(state):
    encrypted_data, persistent_refs = state
    key = getUtility(IKeyHolder).key
    service = getUtility(IEncryptionService)
    stringified_data = service.decrypt(key, encrypted_data)
    return unpickle_nonpersistent(stringified_data, persistent_refs)


def pickle_nonpersistent(state):
    buf = cStringIO.StringIO()
    persistent_refs = []
    cache = {}
    def persistent_id(obj):
        # this should probably handle the same kinds of objects that
        # ZODB.serialize.ObjectWriter.persistent_id does.
        if not isinstance(obj, (persistent.Persistent, type,
                                persistent.wref.WeakRef)):
            # Not persistent, pickle normally
            return None
        # Otherwise let ZODB (instead of our local pickler) handle these.
        idx = cache.get(id(obj), None)
        if idx is None:
            idx = cache[id(obj)] = len(persistent_refs)
            persistent_refs.append(obj)
        return idx
    pickler = cPickle.Pickler(buf, 2)
    pickler.persistent_id = persistent_id
    pickler.dump(state)
    return buf.getvalue(), persistent_refs


def unpickle_nonpersistent(data, persistent_refs):
    buf = cStringIO.StringIO(data)
    def persistent_load(ref):
        return persistent_refs[ref]
    unpickler = cPickle.Unpickler(buf)
    unpickler.persistent_load = persistent_load
    return unpickler.load()


def convert_object_to_encrypted(obj):
    """Convert a Persistent object to EncryptedPersistent.

    ``obj`` should be a persistent ghost of a class that used to subclass
    Persistent and now subclasses EncryptedPersistent.

    DO NOT USE THIS METHOD FOR REAL DATA without fully understanding all
    the implications (existing non-persistent bits will continue to be
    stored on the disk).  Also, this function uses black magic.
    """

    # Black magic happens here
    real_setstate = obj.__class__.__setstate__
    obj.__class__.__setstate__ = persistent.Persistent.__setstate__
    obj._p_activate()
    obj.__class__.__setstate__ = real_setstate
    obj._p_changed = True




##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Implementation of Key Management Facility

$Id: facility.py 90845 2008-09-04 21:06:25Z mgedmin $
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"
import M2Crypto
import datetime
import md5
import persistent
import time
import zope.interface
import zope.location
from z3c.rest import client
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.app.container import btree
from zope.dublincore import property
from zope.schema.fieldproperty import FieldProperty
from keas.kmi import interfaces

class Key(zope.location.Location, persistent.Persistent):
    zope.interface.implements(interfaces.IKey, IAttributeAnnotatable)

    created = property.DCProperty('created')
    creator = property.DCProperty('creator')
    key = FieldProperty(interfaces.IKey['key'])

    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return '<%s %r>' %(self.__class__.__name__, self.key)


class EncryptionService(object):

    cipher = 'aes_256_cbc'

    # Note: decryption fails if you use an empty initialization vector; but it
    # only fails when you restart the Python process.  The length of the
    # initialization vector is assumed to be 16 bytes because that's what
    #   openssl aes-256-cbc -nosalt -P -k 'a'
    # prints if you execute it on the command line
    initializationVector = '0123456789ABCDEF'

    def encrypt(self, key, data):
        """See interfaces.IEncryptionService"""
        # 1. Extract the encryption key
        encryptionKey = self.getEncryptionKey(key)
        # 2. Create a cipher object
        cipher = M2Crypto.EVP.Cipher(
            self.cipher, encryptionKey, self.initializationVector, 1)
        # 3. Feed the data to the cipher
        encrypted = cipher.update(data)
        encrypted += cipher.final()
        # 4. Return encrypted data
        return encrypted

    def decrypt(self, key, data):
        """See interfaces.IEncryptionService"""
        # 1. Extract the encryption key
        encryptionKey = self.getEncryptionKey(key)
        # 2. Create a cipher object
        cipher = M2Crypto.EVP.Cipher(
            self.cipher, encryptionKey, self.initializationVector, 0)
        # 3. Feed the data to the cipher
        decrypted = cipher.update(data)
        decrypted += cipher.final()
        # 4. Return encrypted data
        return decrypted


class KeyManagementFacility(EncryptionService, btree.BTreeContainer):
    zope.interface.implements(interfaces.IExtendedKeyManagementFacility)

    rsaKeyLength = 512 # The length of the key encrypting key
    rsaKeyExponent = 161 # Should be sufficiently high and non-symmetric
    rsaPassphrase = 'key management facility'
    rsaPadding = M2Crypto.RSA.pkcs1_padding

    keyLength = rsaKeyLength/16

    def generate(self):
        """See interfaces.IKeyGenerationService"""
        # 1. Generate the private/public RSA key encrypting key
        rsa = M2Crypto.RSA.gen_key(
            self.rsaKeyLength, self.rsaKeyExponent,
            lambda x: None)
        # 2. Extract the private key from the RSA object
        buf = M2Crypto.BIO.MemoryBuffer('')
        rsa.save_key_bio(buf, None, lambda x: self.rsaPassphrase)
        privateKey = buf.getvalue()
        # 3. Generate the encryption key
        key = M2Crypto.Rand.rand_bytes(self.keyLength)
        # 4. Create the lookup key in the container
        hash = md5.new()
        hash.update(privateKey)
        # 5. Save the encryption key
        encryptedKey = rsa.public_encrypt(key, self.rsaPadding)
        self[hash.hexdigest()] = Key(encryptedKey)
        # 6. Return the private key encrypting key
        return privateKey

    def getEncryptionKey(self, key):
        """Given the key encrypting key, get the encryption key."""
        # 1. Create the lookup key in the container
        hash = md5.new()
        hash.update(key)
        # 2. Extract the encrypted encryption key
        encryptedKey = self[hash.hexdigest()].key
        # 3. Decrypt the key.
        rsa = M2Crypto.RSA.load_key_string(key)
        decryptedKey = rsa.private_decrypt(encryptedKey, self.rsaPadding)
        # 4. Return decrypted encryption key
        return decryptedKey

    def __repr__(self):
        return '<%s (%i)>' %(self.__class__.__name__, len(self))


class LocalKeyManagementFacility(EncryptionService):
    """A local facility that requests keys from the master facility."""
    zope.interface.implements(interfaces.IKeyManagementFacility)

    timeout = 3600
    clientClass = client.RESTClient

    def __init__(self, url):
        self.url = url
        self._cache = {}

    def generate(self):
        """See interfaces.IKeyGenerationService"""
        client = self.clientClass(self.url)
        client.post('/new')
        return client.contents

    def getEncryptionKey(self, key):
        """Given the key encrypting key, get the encryption key."""
        if (key in self._cache and
            self._cache[key][0] + self.timeout > time.time()):
            return self._cache[key][1]
        client = self.clientClass(self.url)
        client.post('/key', key, headers={'content-type': 'text/plain'})
        encryptionKey = client.contents
        self._cache[key] = (time.time(), encryptionKey)
        return encryptionKey

    def __repr__(self):
        return '<%s %r>' %(self.__class__.__name__, self.url)



##############################################################################
#
# Copyright (c) 2008 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST-API to master key management facility

$Id: rest.py 90845 2008-09-04 21:06:25Z mgedmin $
"""
from zope.publisher.browser import BrowserPage

class RestView(BrowserPage):

    def __call__(self):
        method = self.request.method
        if not hasattr(self, method):
            self.request.response.setStatus(405)
            return 'Method not allowed.'
        return getattr(self, method)()

class StatusView(RestView):

    def GET(self):
        self.request.response.setHeader('content-type', 'text/plain')
        return 'KMS server holding %d keys' % len(self.context)

class NewView(RestView):

    def POST(self):
        self.request.response.setHeader('content-type', 'text/plain')
        return self.context.generate()

class KeyView(RestView):

    def POST(self):
        stream = self.request.bodyStream.getCacheStream()
        stream.seek(0)
        key = stream.read()
        self.request.response.setHeader('content-type', 'text/plain')
        try:
            return self.context.getEncryptionKey(key)
        except KeyError:
            self.request.response.setStatus(404)
            return 'Key not found'




##############################################################################
#
# Copyright (c) 2008 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: tests.py 90781 2008-09-04 05:59:31Z srichter $
"""
import unittest

import transaction
from zope.testing import doctest
from zope.app.testing import setup
from zope.component import provideUtility

from keas.kmi.testing import TestingKeyManagementFacility
from keas.kmi.interfaces import IKeyManagementFacility


def setUpPersistent(test):
    setup.setUpTestAsModule(test, name='keas.kmi.tests.doctestfile')
    setup.placelessSetUp()
    provideUtility(TestingKeyManagementFacility(),
                   provides=IKeyManagementFacility)


def tearDownPersistent(test):
    transaction.abort()
    setup.placelessTearDown()
    setup.tearDownTestAsModule(test)


def test_suite():
    return unittest.TestSuite([
        doctest.DocFileSuite(
            'README.txt',
            setUp=setup.placelessSetUp, tearDown=setup.placelessTearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS),
        doctest.DocFileSuite(
            'persistent.txt',
            setUp=setUpPersistent, tearDown=tearDownPersistent,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS),
    ])



##############################################################################
#
# Copyright (c) 2008 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: testing.py 90845 2008-09-04 21:06:25Z mgedmin $
"""
import md5
import cStringIO
from zope.publisher import browser
from zope.interface import implements
from keas.kmi import facility, rest, interfaces

KeyEncyptingKey = '''-----BEGIN RSA PRIVATE KEY-----
MIIBOAIBAAJBAL+VS9lDsS9XOaeJppfK9lhxKMRFdcg50MR3aJEQK9rvDEqNwBS9
rQlU/x/NWxG0vvFCnrDn7VvQN+syb3+a0DMCAgChAkAzKw3lwPxw0VVccq1J7qeO
4DXR1iEMIoWruiCyq0aLkHnQzrZpaHnd4w+JNKIGOVDEWItf3iZNMXkoqj2hoPmp
AiEA5kWTkYrfH+glJUfV/GvU6jcPSNctcJCnqTfMjQU0KEUCIQDU/R3iz5Lojw1S
R1v6C5gdY/mrQydegHVGFS/p276KFwIgDk124UnRb7IyAlDI6xteUVSDVZ4Pivd+
fP6yEkylbQkCIDo1Ri4VvzRtDsVkmFdqhcucHhIu+UTCHN6uVjy7QmIpAiB7Gl9m
F4a4UlXVivb82J7ew3ZABnFIC9Q+dHW7WeZhxg==
-----END RSA PRIVATE KEY-----
'''

EncryptedEncryptionKey = (
    '\xbc\x08\xdbo\x04\xe3\xc7G\x13\xd3\x86\x92\xfa\xe8i>,+\xda\xf8/B2]s\xd4'
    '\x10}[\xfd\x19\x98\xb1\xfa*V~U\xdf\t\x02\x01\xa6\xa8\xae\x8b\x8cm\xd9n'
    '\xd5\x83\xa1%k\x16lfuY\\q\x8c\x8b')

class FakeRESTClient(object):

    context = None

    def __init__(self, url):
        self.url = url

    def post(self, url, data=None, headers={}):
        io = cStringIO.StringIO(data) if data else None
        request = browser.TestRequest(io)
        request.method = 'POST'
        if url == '/new':
            klass = rest.NewView
        elif url == '/key':
            if headers.get('content-type') != 'text/plain':
                # ensure we don't trip on
                # http://trac.pythonpaste.org/pythonpaste/ticket/294
                raise ValueError('bad content type')
            klass = rest.KeyView
        else:
            raise ValueError(url)

        view = klass(self.context, request)
        self.contents = view()


def setupRestApi(localFacility, masterFacility):
    MyFakeRESTClient = type(
        'FakeRESTClient', (FakeRESTClient,), {'context': masterFacility})
    localFacility.clientClass = MyFakeRESTClient


class TestingKeyManagementFacility(facility.KeyManagementFacility):

    def __init__(self):
        super(TestingKeyManagementFacility, self).__init__()
        hash = md5.new()
        hash.update(KeyEncyptingKey)
        md5Key = hash.hexdigest()
        self[md5Key] = facility.Key(EncryptedEncryptionKey)

    def generate(self):
        return KeyEncyptingKey


class TestingKeyHolder(object):
    implements(interfaces.IKeyHolder)
    key = KeyEncyptingKey




##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: interfaces.py 90781 2008-09-04 05:59:31Z srichter $
"""
__docformat__ = "reStructuredText"
import zope.interface
import zope.schema
from zope.app.container import interfaces


class IKey(zope.interface.Interface):
    """Encryption Key"""

    created = zope.schema.Datetime(
        title=u'Creation Date/Time',
        description=u'The date/time the key pair was created.',
        required=True)

    creator = zope.schema.TextLine(
        title=u'Creator',
        description=u'The principal/user that requested the new key.',
        required=True)

    key = zope.schema.Bytes(
        title=u'Key',
        description=u'The key used to encrypt and decrypt the data.',
        required=True)


class IEncryptionService(zope.interface.Interface):
    """Utility providing encryption mechanism"""

    def encrypt(key, data):
        """Returns the encrypted data"""

    def decrypt(key, data):
        """Returns the decrypted data"""


class IKeyGenerationService(zope.interface.Interface):
    """A component that can generate a key encryption pair."""

    def generate():
        """Generate a new set of keys.

        Returns the private key encrypting key.
        """

class IKeyManagementFacility(IEncryptionService, IKeyGenerationService):
    """Key Management Facility

    A key management facility provides several key management services.
    """

class IExtendedKeyManagementFacility(IKeyManagementFacility,
                                     interfaces.IContainer):
    """Extended Key Management Facility.

    This facility also also the management of the keys via Python's mapping
    API.
    """


class IKeyHolder(zope.interface.Interface):
    """Key Holder

    A key holder stores a single key.
    """

    key = zope.schema.Bytes(
        title=u'Key',
        description=u'The key used to encrypt and decrypt the data.',
        required=True)

